@@ -6,6 +6,8 @@
 # Modified Work Copyright (c) 2019 Bradley Beach, esp32spi_mqtt
 # Modified Work Copyright (c) 2012-2019 Roger Light and others, Paho MQTT Python
 
+# pylint: disable=too-many-lines
+
 """
 `adafruit_minimqtt`
 ================================================================================
@@ -32,7 +34,7 @@ from random import randint
 from micropython import const
 from .matcher import MQTTMatcher
 
-__version__ = "5.3.3"
+__version__ = "5.4.2"
 __repo__ = "https://github.com/adafruit/Adafruit_CircuitPython_MiniMQTT.git"
 
 # Client-specific variables
@@ -126,9 +128,12 @@ class MQTT:
     :param str client_id: Optional client identifier, defaults to a unique, generated string.
     :param bool is_ssl: Sets a secure or insecure connection with the broker.
     :param int keep_alive: KeepAlive interval between the broker and the MiniMQTT client.
+    :param int recv_timeout: receive timeout, in seconds.
     :param socket socket_pool: A pool of socket resources available for the given radio.
     :param ssl_context: SSL context for long-lived SSL connections.
     :param bool use_binary_mode: Messages are passed as bytearray instead of string to callbacks.
+    :param int socket_timeout: How often to check socket state for read/write/connect operations,
+        in seconds.
 
     """
 
@@ -142,9 +147,11 @@ class MQTT:
         client_id=None,
         is_ssl=True,
         keep_alive=60,
+        recv_timeout=10,
         socket_pool=None,
         ssl_context=None,
         use_binary_mode=False,
+        socket_timeout=1,
     ):
 
         self._socket_pool = socket_pool
@@ -153,6 +160,13 @@ class MQTT:
         self._backwards_compatible_sock = False
         self._use_binary_mode = use_binary_mode
 
+        if recv_timeout <= socket_timeout:
+            raise MMQTTException(
+                "recv_timeout must be strictly greater than socket_timeout"
+            )
+        self._socket_timeout = socket_timeout
+        self._recv_timeout = recv_timeout
+
         self.keep_alive = keep_alive
         self._user_data = None
         self._is_connected = False
@@ -214,7 +228,7 @@ class MQTT:
 
         :param str host: Desired broker hostname
         :param int port: Desired broker port
-        :param int timeout: Desired socket timeout
+        :param int timeout: Desired socket timeout, in seconds
         """
         # For reconnections - check if we're using a socket already and close it
         if self._sock:
@@ -252,6 +266,7 @@ class MQTT:
 
         sock = None
         retry_count = 0
+        last_exception = None
         while retry_count < 5 and sock is None:
             retry_count += 1
 
@@ -268,14 +283,19 @@ class MQTT:
 
             try:
                 sock.connect((connect_host, port))
-            except MemoryError:
+            except MemoryError as exc:
                 sock.close()
                 sock = None
-            except OSError:
+                last_exception = exc
+            except OSError as exc:
                 sock.close()
                 sock = None
+                last_exception = exc
 
         if sock is None:
+            if last_exception:
+                raise RuntimeError("Repeated socket failures") from last_exception
+
             raise RuntimeError("Repeated socket failures")
 
         self._backwards_compatible_sock = not hasattr(sock, "recv_into")
@@ -444,7 +464,9 @@ class MQTT:
             self.logger.debug("Attempting to establish MQTT connection...")
 
         # Get a new socket
-        self._sock = self._get_connect_socket(self.broker, self.port)
+        self._sock = self._get_connect_socket(
+            self.broker, self.port, timeout=self._socket_timeout
+        )
 
         # Fixed Header
         fixed_header = bytearray([0x10])
@@ -514,6 +536,7 @@ class MQTT:
             self._send_str(self._password)
         if self.logger is not None:
             self.logger.debug("Receiving CONNACK packet from broker")
+        stamp = time.monotonic()
         while True:
             op = self._wait_for_msg()
             if op == 32:
@@ -527,6 +550,12 @@ class MQTT:
                     self.on_connect(self, self._user_data, result, rc[2])
                 return result
 
+            if op is None:
+                if time.monotonic() - stamp > self._recv_timeout:
+                    raise MMQTTException(
+                        f"No data received from broker for {self._recv_timeout} seconds."
+                    )
+
     def disconnect(self):
         """Disconnects the MiniMQTT client from the MQTT broker."""
         self.is_connected()
@@ -637,6 +666,7 @@ class MQTT:
         if qos == 0 and self.on_publish is not None:
             self.on_publish(self, self._user_data, topic, self._pid)
         if qos == 1:
+            stamp = time.monotonic()
             while True:
                 op = self._wait_for_msg()
                 if op == 0x40:
@@ -649,6 +679,12 @@ class MQTT:
                             self.on_publish(self, self._user_data, topic, rcv_pid)
                         return
 
+                if op is None:
+                    if time.monotonic() - stamp > self._recv_timeout:
+                        raise MMQTTException(
+                            f"No data received from broker for {self._recv_timeout} seconds."
+                        )
+
     def subscribe(self, topic, qos=0):
         """Subscribes to a topic on the MQTT Broker.
         This method can subscribe to one topics or multiple topics.
@@ -697,6 +733,7 @@ class MQTT:
             for t, q in topics:
                 self.logger.debug("SUBSCRIBING to topic %s with QoS %d", t, q)
         self._sock.send(packet)
+        stamp = time.monotonic()
         while True:
             op = self._wait_for_msg()
             if op == 0x90:
@@ -710,6 +747,12 @@ class MQTT:
                     self._subscribed_topics.append(t)
                 return
 
+            if op is None:
+                if time.monotonic() - stamp > self._recv_timeout:
+                    raise MMQTTException(
+                        f"No data received from broker for {self._recv_timeout} seconds."
+                    )
+
     def unsubscribe(self, topic):
         """Unsubscribes from a MQTT topic.
 
@@ -747,6 +790,7 @@ class MQTT:
         if self.logger is not None:
             self.logger.debug("Waiting for UNSUBACK...")
         while True:
+            stamp = time.monotonic()
             op = self._wait_for_msg()
             if op == 176:
                 rc = self._sock_exact_recv(3)
@@ -759,6 +803,12 @@ class MQTT:
                     self._subscribed_topics.remove(t)
                 return
 
+            if op is None:
+                if time.monotonic() - stamp > self._recv_timeout:
+                    raise MMQTTException(
+                        f"No data received from broker for {self._recv_timeout} seconds."
+                    )
+
     def reconnect(self, resub_topics=True):
         """Attempts to reconnect to the MQTT broker.
 
